/*
  DDS Controler
  for ATtiny2313 (8.294400MHz)

  T. Nakagawa

  2010-02-22
*/


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "14seglcd.h"


#define FREQUENCY 180000000	/* (*NbNZ)̎g */
#define OSCCALVAL 110	/* NbN8294400ɂNbN␳l */
#define KEYRPT_START 60	/* L[s[gJn鎞(*10ms) */
#define KEYRPT_INTVL 8	/* L[s[g̊Ԋu(*10ms) */
#define FREQ_MAX 70000000	/* őݒ\g */
#define DDS_DDR DDRD
#define DDS_PORT PORTD
#define DDS_BIT_WCLK PD3
#define DDS_BIT_FQUD PD4
#define DDS_BIT_DATA PD5
#define SW_PIN PINB
#define SW_BIT_LT PB3
#define SW_BIT_RT PB2
#define SW_BIT_UP PB1
#define SW_BIT_DN PB0


extern uint32_t division(uint32_t dividend, uint32_t divisor);


static uint8_t mode = 0;	/* ݂̃[h(0:}jAC!=0:[g) */
volatile static uint8_t newmode;
static uint8_t crsr = 0;	/* J[\ʒu(0:1ځ`7:8ځC8:[h) */
volatile static uint8_t newcrsr;
static uint32_t freq = 0;	/* ݂̎g */
volatile static uint32_t newfreq;
static uint32_t uart_data;	/* ݂܂łɓǂݍUARTf[^ */
static uint8_t uart_cnt = 0;	/* ݂܂łɓǂݍUARTf[^ */


const PROGMEM uint32_t updn[8] = {	/* up/downXCb`̑ */
  1,
  10,
  100,
  1000,
  10000,
  100000,
  1000000,
  10000000,
};


/*
  DDS̎gݒ
*/
static void dds_set(void) {
  uint32_t d;
  uint8_t i;

  d = division(freq, FREQUENCY);
  for (i = 0; i < 40; i++) {
    if (i == 32) {
      d = 0x00000001;
    }
    if ((d & 0x00000001)) {
      DDS_PORT |= _BV(DDS_BIT_DATA);
    } else {
      DDS_PORT &= ~_BV(DDS_BIT_DATA);
    }
    d >>= 1;
    DDS_PORT |= _BV(DDS_BIT_WCLK);
    DDS_PORT &= ~_BV(DDS_BIT_WCLK);
  }
  DDS_PORT |= _BV(DDS_BIT_FQUD);
  DDS_PORT &= ~_BV(DDS_BIT_FQUD);

  return;
}


/*
  DDS̏
*/
static void dds_init(void) {
  DDS_DDR |= (_BV(DDS_BIT_WCLK) | _BV(DDS_BIT_FQUD) | _BV(DDS_BIT_DATA));
  DDS_PORT |= _BV(DDS_BIT_WCLK);
  DDS_PORT &= ~_BV(DDS_BIT_WCLK);
  DDS_PORT |= _BV(DDS_BIT_FQUD);
  DDS_PORT &= ~_BV(DDS_BIT_FQUD);

  dds_set();

  return;
}


/*
  LCDւ̕\
*/
static void display(void) {
  int8_t i;
  uint32_t n;
  uint8_t m;

  n = freq;
  for (i = 7; i >= 0; i--) {
    m = n % 10;
    lcd_buf[i] = m;
    n /= 10;
    if (n == 0) break;
  }
  for (i--; i >= 0; i--) {
    lcd_buf[i] = LCD_SPACE;
  }
  lcd_buf[8] = mode ? 0x80 : 0x20;
  if (crsr == 8) {
    lcd_buf[8] |= 0x40;
  } else {
    lcd_buf[7 - crsr] += LCD_CURSOR;
  }

  lcd_refresh();

  return;
}


int main(void) {
  /* CPUNbN(8294400Hz) */
  OSCCAL = OSCCALVAL;

  /* |[g(gps̓vAbv) */
  PORTB = (_BV(PB0) | _BV(PB1) | _BV(PB2) | _BV(PB3) | _BV(PB4));
  PORTD = (_BV(PD0) | _BV(PD1) | _BV(PD2) | _BV(PD6));

  /* UART(115200baud) */
  UCSRA |= _BV(U2X);
  UCSRB |= (_BV(RXCIE) | _BV(RXEN));
  UCSRC |= (_BV(UCSZ1) | _BV(UCSZ0));
  UBRRL = 8;

  /* ^C}[(10msɊ荞) */
  TCCR1B |= (_BV(WGM12) | _BV(CS12) | _BV(CS10));
  TIMSK |= _BV(OCIE1A);
  OCR1A = 81 - 1;

  /* DDS */
  dds_init();

  /* LCD */
  lcd_init();
  display();

  sei();

  for (; ; ) {
    if (newfreq != freq) {
      cli();
      freq = newfreq;
      sei();
      dds_set();
      display();
    } else if (newcrsr != crsr || newmode != mode) {
      crsr = newcrsr;
      mode = newmode;
      display();
    } 
  }

  return 0;
}


/*
  UARTM荞
*/
ISR (USART_RX_vect) {
  uint8_t buf;

  buf = UDR;

  if (mode) {
    uart_data <<= 8;
    uart_data |= (uint32_t)buf;
    uart_cnt++;
    if (uart_cnt == 4) {
      newfreq = uart_data;
      if (newfreq > FREQ_MAX) newfreq = FREQ_MAX;
      uart_cnt = 0;
    }
  }
}


/*
  ^C}[荞(10ms)
*/
ISR (TIMER1_COMPA_vect) {
  static uint8_t prv = 0x00, cnt = 0;
  uint8_t cur, stt;

  cur = ~SW_PIN;	/* vAbvĂ̂Ŕ] */
  stt = ((cur ^ prv) & cur);	/* OFF->ONɂȂXCb`̃rbg𗧂Ă */

  if (crsr == 8) {	/* [hؑ */
    if ((stt & (_BV(SW_BIT_UP) | _BV(SW_BIT_DN)))) {
      newmode = ~newmode;
      uart_cnt = 0;
    }
  } else {	/* g̑ */
    if ((cur & (_BV(SW_BIT_UP) | _BV(SW_BIT_DN)))) {
      if (cnt == 0 || cnt == KEYRPT_START) {
        if ((cur & _BV(SW_BIT_UP))) {
          newfreq = freq + pgm_read_dword(&updn[crsr]);
          if (newfreq > FREQ_MAX) newfreq = FREQ_MAX;
        } else {
          newfreq = freq - pgm_read_dword(&updn[crsr]);
          if (newfreq > FREQ_MAX) newfreq = 0;
        }
        if (cnt == KEYRPT_START) cnt -= KEYRPT_INTVL;
      }
      cnt++;
    } else {
      cnt = 0;
    }
  }

  /* J[\̈ړ */
  if (!mode) {
    if ((stt & _BV(SW_BIT_LT)) && crsr != 8) {
      newcrsr = crsr + 1;
    } else if ((stt & _BV(SW_BIT_RT)) && crsr != 0) {
      newcrsr = crsr - 1;
    }
  }

  prv = cur;

  return;
}
